/* Terraform - (C) 1997-1999 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <math.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "terraform.h"
#include "TFWindowHandler.h"
#include "FileIO.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"
#include "GuiColormap.h"
#include "GuiDialogPreview.h"
#include "GuiDialogReallyQuit.h"
#include "HeightFieldExport.h"
#include "HeightFieldGenRandom.h"
#include "HeightFieldIO.h"
#include "MenuDefs.h"
#include "TFDialogContourLines.h"
#include "TFDialogCraters.h"
#include "TFDialogErode.h"
#include "TFDialogFill.h"
#include "TFDialogFlowmap.h"
#include "TFDialogGaussianHill.h"
#include "TFDialogFold.h"
#include "TFDialogGenSSynth.h"
#include "TFDialogGenSubdiv.h"
#include "TFDialogLinearScale.h"
#include "TFDialogMerge.h"
#include "TFDialogMirror.h"
#include "TFDialogOptions.h"
#include "TFDialogPPov.h"
#include "TFDialogRescale.h"
#include "TFDialogRotate.h"
#include "TFDialogRoughSmooth.h"
#include "TFDialogTerrace.h"
#include "TFDialogTransform.h"
#include "TFGui.h"					// circular dependency
#include "TFOptions.h"
#include "TFPreviewDialog.h"
#include "Timer.h"
#include "MathRandom.h"
#include "agename.h"

#ifdef HAVE_IMLIB
#include "GuiDialogAboutDissolve.h"
#endif


extern int readhf (HeightField **HF, char *fname, bool haveGUI);

static int		cs_fileselMode;	// flag for file selection pointer


#define LOAD		0
#define SAVE		1
#define BANDSIZE	15


/*
 *  constructor: initialize everything 
 */
TFWindowHandler::TFWindowHandler (FlexArray *wList) 
		: TFWindowBase (wList)
		// default constructor is OK for TFWindow
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ TFWindowHandler\n");

	d_HF = NULL;
	d_HFD = NULL;
	d_fileSelection = NULL;
	d_genSSynthDialog=d_genSubdivDialog=d_placePovDialog=NULL;
	d_contLinesDialog=d_erodeDialog=d_flowmapDialog=d_rescaleDialog = NULL;
	d_cratersDialog=d_fillDialog=d_foldDialog=d_gaussHillDialog=
		d_linearScaleDialog=d_mirrorDialog=d_rotateDialog=
		d_roughSmoothDialog=d_terraceDialog=d_transformDialog = NULL;
	d_fastWire = FALSE;
	d_rotate = FALSE;
	d_autoRotate = FALSE;
	
	activateMenus (FALSE);
	gdk_flush ();
}


/*
 *  destructor: clean up 
 */
TFWindowHandler::~TFWindowHandler ()
{
	if (d_HF) 		delete d_HF;
	if (d_HFD) 		delete d_HFD;
	if (d_fileSelection) 	delete d_fileSelection;
	if (d_contLinesDialog)	delete d_contLinesDialog;
	if (d_cratersDialog)	delete d_cratersDialog;
	if (d_erodeDialog) 	delete d_erodeDialog;
	if (d_flowmapDialog) 	delete d_flowmapDialog;
	if (d_genSSynthDialog) 	delete d_genSSynthDialog;
	if (d_genSubdivDialog) 	delete d_genSubdivDialog;
	if (d_placePovDialog) 	delete d_placePovDialog;
	if (d_rescaleDialog) 	delete d_rescaleDialog;
	if (d_fillDialog) 	delete d_fillDialog;
	if (d_foldDialog) 	delete d_foldDialog;
	if (d_linearScaleDialog) delete d_linearScaleDialog;
	if (d_mirrorDialog) 	delete d_mirrorDialog;
	if (d_roughSmoothDialog) delete d_roughSmoothDialog;
	if (d_terraceDialog)	delete d_terraceDialog;
	if (d_transformDialog)	delete d_transformDialog;
	if (d_previewDialog)	delete d_previewDialog;
	if (d_rotateDialog) 	delete d_rotateDialog;
	
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- TFWindowHandler\n");
}


gint TFWindowHandler::idleRedrawCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Idle Redraw ...\n");
	drawHeightField ();
	return 0;
}


/*
 *  find TFWin: loop through TFWindowBase::s_winList and find the desired HF 
 */
gint TFWindowHandler::waitForChildrenCallback ()
{
	//printf ("Waiting for children\n");

	int	status;
	pid_t	wpid;

	wpid = waitpid (-1, &status, WNOHANG);
	if (wpid > 0)
		return 0;

	return 1;
}


gint TFWindowHandler::autoRotateCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Auto Rotate ...\n");

	if (d_HFD->rotateFastView (0.0, (M_PI/180)*1, 0.0))
		{
		// !!!!!!!!! set incorrectly selected check box back to FALSE
		d_autoRotate = FALSE;
		}
	else
		drawHeightField ();

	return d_autoRotate;
}


gint TFWindowHandler::calibrateAutoRotate (int nSamples)
{
	Timer		timer;
	long		eusec;		// elapsed microseconds
	struct timeval  *etime;		// elapsed time

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Calibrating AutoRotate ...");

	timer.start ();
	for (int i=0; i<nSamples; i++)
		{
		autoRotateCallback ();
		gdk_flush ();
		}
	timer.stop ();
	etime = timer.getElapsedTime ();
	eusec = etime->tv_usec + etime->tv_sec*1000000;
	eusec /= nSamples;

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW,  " done\n");

	return eusec;
}


/*
 *  set the Height Field we draw 
 */
void TFWindowHandler::setHeightField (HeightField *HF)
{
	int	x, y;
	bool	upd = FALSE;

	SanityCheck::bailout ((!HF), "HF==NULL", "TFWindowHandler::setHeightField");

	if (!d_HF)
		upd = TRUE;
		
	if (d_HF && d_HF!=HF)
		delete d_HF;
	d_HF=HF;

	if (d_HFD)
		delete d_HFD;
	d_HFD=new HeightFieldDraw (d_HF, this);
	d_HFD->setScale (1.0, &x, &y);
	if (upd)
		d_HFD->updateParams ();

	if (this->d_useLinearColormap)
		d_HFD->setColormap (this->d_linearColorMap);
	else
		d_HFD->setColormap (this->d_bandColorMap);

	if (GlobalTrace::isSet(GlobalTrace::TRACE_DEBUG))
		d_HFD->printSettings();

	//d_HFD->draw ();
	d_win.set_title (d_HF->getName());
	connect_to_method (Gtk_Main::timeout(50), this, &TFWindowHandler::idleRedrawCallback);
	activateMenus (TRUE);
}


/*
 *  drawHeightField: draw and then reset fore- and background colors
 */
void TFWindowHandler::drawHeightField ()
{
	SanityCheck::bailout ((!d_HF), "d_HF==NULL", "TFWindowHandler::drawHeightField");
	SanityCheck::bailout ((!d_HFD), "d_HFD==NULL", "TFWindowHandler::drawHeightField");

	d_HFD->draw ();
	this->setColors (&(this->d_linearColorMap->d_black));
}



/*
 *  fileMenuCallback: process a menu choice in the file menu
 */
void TFWindowHandler::fileMenuCallback (string m)
{
	int		rc=0;
	char		*menuitem = const_cast<char*>(m.c_str()),
			*aboutText[] = { "brought to you by", 
					"Robert_Gasch@peoplesoft.com", };

	if (!strcmp (menuitem, MENU_FILE_NEW_SYNTHESIS))
		{
		if (!d_genSSynthDialog)
			d_genSSynthDialog = new TFDialogGenSSynth ();
		else
			{
			TFDialogGenSSynth	*TFDGen;

			TFDGen = dynamic_cast<TFDialogGenSSynth*>(d_genSSynthDialog);
			TFDGen->reset ();
			TFDGen->set_title ("New Height Field");
			}
		
		d_genSSynthDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_NEW_SUBDIVIDE))
		{
		if (!d_genSubdivDialog)
			d_genSubdivDialog = new TFDialogGenSubdiv ();
		else
			{
			TFDialogGenSubdiv 	*TFDGen;

			TFDGen = dynamic_cast<TFDialogGenSubdiv*>(d_genSubdivDialog);
			TFDGen->reset ();
			TFDGen->set_title ("New Height Field");
			}
		
		d_genSubdivDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_NEW_RANDOM))
		{
		if (!d_HF)
			d_HF = new HeightField ();

		HeightFieldGenRandom::generate (d_HF, this->width());
		setHeightField (d_HF);
		//d_HFD->draw ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_OPEN))
		{
		cs_fileselMode=LOAD;			// 0 = load, 1 = save
		fileSelection ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_MERGE))
		{
		TFDialogMerge 	*mergeDialog = TFGui::getMergeDialog ();
		mergeDialog->rebuildCList ();
		mergeDialog->initHeightFields (this);
		mergeDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_RELOAD))
		{
		if (d_HF && !d_HF->isSaved())
			{
			if (!TFWindowBase::s_quitDialog)
				TFWindowBase::s_quitDialog = new GuiDialogReallyQuit ("", "");
			else
				TFWindowBase::s_quitDialogConnection.disconnect ();

			TFWindowBase::s_quitDialog->set_title (d_HF->getName());
			TFWindowBase::s_quitDialog->setLabel ((char *)"Unsaved Changes! Reload Anyways?");
			TFWindowBase::s_quitDialogConnection = connect_to_method (
				TFWindowBase::s_quitDialog->getYesButton()->clicked, 
				this, &TFWindowHandler::reloadCallback);
			TFWindowBase::s_quitDialog->show ();
			}
		else
		if (d_HF)
			reloadCallback ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_SAVE))
		{
		if (d_HF)
			{
			HeightFieldIO		*HFIO;

			HFIO = new HeightFieldIO (d_HF->getName());
			rc = HFIO->write (d_HF);
			}
		}
	else
	if (!strcmp (menuitem, MENU_FILE_SAVEAS))
		{
		if (d_HF)
			{
			cs_fileselMode=SAVE;			// 0 = load, 1 = save
			fileSelection ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_FILE_CLOSE))
		{
		if (d_HF && !d_HF->isSaved())
			{
			if (!TFWindowBase::s_quitDialog)
				TFWindowBase::s_quitDialog = new GuiDialogReallyQuit ("", "");
			else
				TFWindowBase::s_quitDialogConnection.disconnect ();

			TFWindowBase::s_quitDialog->set_title (d_HF->getName());
			TFWindowBase::s_quitDialog->setLabel ("Unsaved Changes! Close Anyways?");
			TFWindowBase::s_quitDialogConnection = connect_to_method 
				(TFWindowBase::s_quitDialog->getYesButton()->clicked, 
				 this, &TFWindowHandler::closeCallback);
			TFWindowBase::s_quitDialog->show ();
			}
		else
		if (d_HF)
			closeCallback ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_EXPORT_POVRAY))
		{
		HeightFieldExport	*HFE = new HeightFieldExport (d_HF);
		float			slvl, yscale;

		if (TFOptions::s_fillSea)
			slvl = TFOptions::s_sealevel;
		else
			slvl = 0;

		if (TFOptions::s_POVhalfYscale)
			yscale = TFOptions::s_yscale*.5;
		else
			yscale = TFOptions::s_yscale;

		HFE->renderPOV (TFOptions::s_POVfile, TFOptions::s_POVwidth, 
			yscale, slvl, TFOptions::s_POVkeepFiles);
		delete HFE;
		connect_to_method (Gtk_Main::timeout(1000), this, &TFWindowHandler::waitForChildrenCallback);
		}
	else
	if (!strcmp (menuitem, MENU_FILE_ABOUT))
		{
		if (!TFWindowBase::s_aboutDialog)
			{
#ifdef HAVE_IMLIB
			char	buf[256];
			sprintf (buf, "%s/terraform-logo.jpg", TF_DATADIR);
			FileIO	*fio = new FileIO (buf);
			if (fio->exists())
				TFWindowBase::s_aboutDialog = new GuiDialogAboutDissolve ("About Terraform", buf, aboutText, 2);
			else
				fprintf (stderr, "Error opening %s\n", buf);
			delete fio;
#else
			TFWindowBase::s_aboutDialog = new GuiDialogAbout ("About Terraform", NULL, aboutText, 2);
#endif
			if (TFWindowBase::s_aboutDialog)
				TFWindowBase::s_aboutDialog->buildDialogWindow ();
			}

		if (TFWindowBase::s_aboutDialog)
			TFWindowBase::s_aboutDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_OPTIONS))
		{
		if (d_HF)
			{
			if (!TFWindowBase::s_optionsDialog)
				TFWindowBase::s_optionsDialog = new TFDialogOptions (TFWindowBase::s_winList);
			else
				TFWindowBase::s_optionsDialog->show ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_FILE_QUIT))
		{
		if (!TFGui::allWindowsSaved())
			{
			if (!TFWindowBase::s_quitDialog)
				TFWindowBase::s_quitDialog = new GuiDialogReallyQuit ("", "");
			else
				TFWindowBase::s_quitDialogConnection.disconnect ();

			TFWindowBase::s_quitDialog->set_title ("Really Quit Terraform?");
			TFWindowBase::s_quitDialog->setLabel ("Unsaved Changes! Exit anyway?");
			TFWindowBase::s_quitDialogConnection = connect_to_method (
				TFWindowBase::s_quitDialog->getYesButton()->clicked, 
				this, &TFWindowHandler::destroyCallback);
			TFWindowBase::s_quitDialog->show ();
			}
		else
			destroyCallback ();
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <file> menuitem", "TFWindowHandler::fileMenuCallback");

	if (rc)
		g_print ("Warning: last command returned an Error (%d)\n", rc);
	fflush (stdout);
}


/*
 *  hfMenuCallback: process a menu choice in the height field menu
 */
void TFWindowHandler::hfMenuCallback (string m)
{
	char		*menuitem = const_cast<char*>(m.c_str());

	if (!d_HF)
		return;

	if (!strcmp (menuitem, MENU_HF_CRATERS))
		{
		if (!d_cratersDialog) 
			d_cratersDialog = new TFDialogCraters (d_HF, d_HFD);
		d_cratersDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_FILL))
		{
		if (!d_fillDialog) 
			d_fillDialog = new TFDialogFill (d_HF, d_HFD);
		d_fillDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_FLOWMAP))
		{
		if (!d_flowmapDialog) 
			d_flowmapDialog = new TFDialogFlowmap (d_HF, d_HFD);
		d_flowmapDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_FOLD))
		{
		if (!d_foldDialog)
			d_foldDialog = new TFDialogFold (d_HF, d_HFD);
		d_foldDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_GAUSS_HILL))
		{
		if (!d_gaussHillDialog)
			d_gaussHillDialog= new TFDialogGaussianHill (d_HF, d_HFD);
		d_gaussHillDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_INVERT))
		{
		if (d_HF)
			{
			HeightFieldOps	*HFO;

			HFO = new HeightFieldOps (d_HF);
			HFO->invert ();
			d_HFD->draw ();
			delete HFO;
			}
		}
	else
	if (!strcmp (menuitem, MENU_HF_LINSCALE))
		{
		if (!d_linearScaleDialog) 
			d_linearScaleDialog = new TFDialogLinearScale (d_HF, d_HFD);
		d_linearScaleDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_MIRROR))
		{
		if (!d_mirrorDialog)
			d_mirrorDialog = new TFDialogMirror (d_HF, d_HFD);
		d_mirrorDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_RESCALE))
		{
		if (!d_rescaleDialog) 
			d_rescaleDialog = new TFDialogRescale (d_HF, d_HFD);
		d_rescaleDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_ROTATE))
		{
		if (!d_rotateDialog) 
			d_rotateDialog = new TFDialogRotate (d_HF, d_HFD);
		d_rotateDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_ROUGHSMOOTH))
		{
		if (!d_roughSmoothDialog) 
			d_roughSmoothDialog = new TFDialogRoughSmooth (d_HF, d_HFD);
		d_roughSmoothDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_TERRACE))
		{
		if (!d_terraceDialog)
			d_terraceDialog = new TFDialogTerrace (d_HF, d_HFD);
		d_terraceDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_TRANSFORM))
		{
		if (!d_transformDialog)
			d_transformDialog = new TFDialogTransform (d_HF, d_HFD);
		d_transformDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_CLONE))
		{
		HeightField 	*HF;
		char		*newName = TFGui::getNewName(d_HF->getName());

		HF = new HeightField (d_HF->getCopyOfData(),
			d_HF->getWidth(), d_HF->getHeight(), newName);
		free (newName);

		TFGui::addHeightField (HF);
		}
	else
	if (!strcmp (menuitem, MENU_HF_EXPORT_CONTOUR))
		{
		if (!d_contLinesDialog)
			d_contLinesDialog = new TFDialogContourLines (d_HF, d_HFD);
		d_contLinesDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_CRATERS))
		{
		if (!d_contLinesDialog)
			d_contLinesDialog = new TFDialogContourLines (d_HF, d_HFD);
		d_contLinesDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_ERODE))
		{
		if (!d_erodeDialog)
			d_erodeDialog = new TFDialogErode (d_HF, d_HFD);
		d_erodeDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_PLACEPOV))
		{
		if (!d_placePovDialog)
			d_placePovDialog = new TFDialogPPov (d_HF);
		d_placePovDialog->show ();
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <hf> menuitem", "TFWindowHandler::fileMenuCallback");

	fflush (stdout);
}


/*
 *  viewMenuCallback: process a menu choice in the view menu
 */
void TFWindowHandler::viewMenuCallback (string m)
{
	char		*menuitem = const_cast<char*>(m.c_str());
	long		delay;

	if (!d_HFD)
		return ;

	if (!strcmp (menuitem, MENU_VIEW_FASTWIRE))
		{
		Gtk_Widget		*w;

		if (d_fastWire)
			d_fastWire = FALSE;
		else
			d_fastWire = TRUE;
		w = static_cast<Gtk_Widget*>(this->d_menuHash->lookup(MENU_VIEW_AUTOROTATE));
		w->set_sensitive (d_fastWire);
		d_HFD->setFastWire (d_fastWire);
		if (d_HFD->getMode()==WIRE)
			drawHeightField ();
		}
	else
	if (!strcmp (menuitem, MENU_VIEW_AUTOROTATE))
		{
		if (d_autoRotate)
			d_autoRotate = FALSE;
		else
			{
			d_autoRotate = TRUE;
			delay = (long int)(calibrateAutoRotate(10)/1000*1.1);
			connect_to_method (Gtk_Main::timeout(delay), this, 
					&TFWindowHandler::autoRotateCallback);
			}
		}
	else
	if (!strcmp (menuitem, MENU_VIEW_2DPLANE) ||
	    !strcmp (menuitem, MENU_VIEW_3DWIRE)  ||
	    !strcmp (menuitem, MENU_VIEW_3DHEIGHT)||
	    !strcmp (menuitem, MENU_VIEW_3DLIGHT) )
		{
		if (!d_HFD->setMode (menuitem))
			drawHeightField ();
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <view> menuitem", "TFWindowHandler::fileMenuCallback");

	fflush (stdout);
}

/*
 *  colormapMenuCallback: process a menu choice in the colormap menu
 */
void TFWindowHandler::colormapMenuCallback (string m)
{
	char		*menuitem = const_cast<char*>(m.c_str());

	if (!strcmp (menuitem, MENU_CMAP_GRAYSCALE))
		{
		this->d_useLinearColormap = TRUE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_linearColorMap))
				d_HFD->draw ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_CMAP_COLORBANDS))
		{
		this->d_useLinearColormap = FALSE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_bandColorMap)) 
				d_HFD->draw ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_CMAP_DESERT))
		{
		this->d_useLinearColormap = FALSE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_bandColorMap2))
				d_HFD->draw ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_CMAP_REDHOT))
		{
		this->d_useLinearColormap = TRUE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_linearColorMap2))
				d_HFD->draw ();
			}
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <colormap> menuitem", "TFWindowHandler::fileMenuCallback");

	fflush (stdout);
}


/*
 *  destroyCallback: handle a destroy event 
 */
void TFWindowHandler::destroyCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Entering destroyCallback\n"); 

	closeCallback ();

	if (TFOptions::s_settingsAreChanged)
		{
		TFOptions	*tfOpt = new TFOptions ();

		tfOpt->install ();
		delete tfOpt;
		}

	if (GlobalTrace::isSet(GlobalTrace::TRACE_VERBOSE))
		g_print ("Shutting Down ...\n");
	gtk_exit (0);
	exit (0);
}


/*
 *  closeCallback: delete the height field
 */
void TFWindowHandler::closeCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Entering closeCallback\n"); fflush (stdout);

	if (TFWindowBase::s_quitDialog)
		TFWindowBase::s_quitDialog->hide ();

	if (d_HFD)
		d_HFD->clear ();

	activateMenus (FALSE);

	if (this->d_useLinearColormap)
		this->setColors (&(d_linearColorMap->d_black));
	else
		this->setColors (&(d_bandColorMap->d_black));
	this->drawRectangle (0, 0, this->width(), 
			this->height(), TRUE);
	this->sync ();
	TFGui::scheduleDelete (d_HF);
	connect_to_function (Gtk_Main::timeout(100), &TFGui::cleanup);
}


/*
 *  reloadCallback: process a reload request 
 */
void TFWindowHandler::reloadCallback ()
{
	if (TFWindowBase::s_quitDialog)
		TFWindowBase::s_quitDialog->hide ();

	if (d_HF)
		{
		HeightField		*HF = NULL;

		readhf (&HF, d_HF->getName(), TRUE);
		setHeightField (HF);
		}
}


/*
 *  fileSelectionOK: user pressed OK in file selection dialog
 */
void TFWindowHandler::fileSelectionOK ()
{
	gchar 	*fname;
	int	rc;

	d_fileSelection->hide ();

	fname = const_cast<char*>(d_fileSelection->get_filename().c_str());
	//g_print ("%s\n", fname);

	//create_progress_bar (fname);

	if (cs_fileselMode == LOAD)
		{
		HeightField	*HF=NULL;

		rc = readhf (&HF, fname, TRUE);
		setHeightField (HF);
		}
	else
		{
		HeightFieldIO		*HFIO;

		HFIO = new HeightFieldIO (fname);
                rc = HFIO->write (d_HF);
		delete HFIO;
		}
	if (rc)
		cout << "Warning: last command returned Error " << rc << "\n";

	activateMenus (TRUE);
}


/*
 *  fileSelectionCancel: user pressed Cancel in file selection dialog
 */
void TFWindowHandler::fileSelectionCancel ()
{
	d_fileSelection->set_filename ("");
	d_fileSelection->hide ();
}


/*
 *  fileSelection: open a file selection dialog
 */
void TFWindowHandler::fileSelection ()
{
	if (!d_fileSelection)
		{
		d_fileSelection = new Gtk_FileSelection ("");
		d_fileSelection->set_position (GTK_WIN_POS_MOUSE);

		connect_to_method (d_fileSelection->get_ok_button()->clicked, 
					this, &TFWindowHandler::fileSelectionOK);
		connect_to_method (d_fileSelection->get_cancel_button()->clicked,
					this, &TFWindowHandler::fileSelectionCancel);
		}

	if (cs_fileselMode==LOAD)
		d_fileSelection->set_title ("Load Height Field");
	else
		d_fileSelection->set_title ("Save Height Field");

	d_fileSelection->show ();
}


void TFWindowHandler::checkMenus ()
{
	Gtk_ObjectHandle<Gtk_MenuItem> 	m;
	string				*s;

	//s = new string (MENU_FILE_MERGE);	// this fails because of the _
	s = new string ("/File/Merge");
	m = d_itemFactory->get_menuitem_widget (*s);
	if (s_winList->getSize() > 1)
		m->set_sensitive (TRUE);
	else
		m->set_sensitive (FALSE);
	delete s;

	//s = new string (MENU_FILE_RELOAD);	// this fails because of the _
	s = new string ("/File/Reload");	// this fails because of the _
	m = d_itemFactory->get_menuitem_widget (*s);
	if (d_HF->isLoadedFromFile())
		m->set_sensitive (TRUE);
	else
		m->set_sensitive (FALSE);
	delete s;
}


void TFWindowHandler::activateMenus (bool activate)
{
	Gtk_Widget	**w;
	int		c=0, i; 

	if (SanityCheck::warning((!this->d_menuHash->get_size()), "size==0", "TFWindowHandler::activateMenus"))
		return;
	
	w = new Gtk_Widget*[this->d_menuHash->get_size()];

	if (!activate || (d_HF && d_HF->isLoadedFromFile()))
		w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_RELOAD));
	if (!activate || s_winList->getSize() > 1)
		w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_MERGE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_SAVE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_SAVEAS));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_CLOSE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_EXPORT_POVRAY));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_OPTIONS));
	if (activate)
		w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_FILE_ABOUT));
	//w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_BRANCH));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_CRATERS));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_FILL));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_FLOWMAP));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_FOLD));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_GAUSS_HILL));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_INVERT));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_LINSCALE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_MIRROR));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_RESCALE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_ROTATE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_ROUGHSMOOTH));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_TERRACE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_TRANSFORM));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_EXPORT_CONTOUR));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_CRATERS));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_ERODE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_HF_PLACEPOV));
	//w[c++] = static_cast<Gtk_Widgt*> (this->d_menuHash->lookup (MENU_VIEW_BRANCH));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_2DPLANE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_3DWIRE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_3DHEIGHT));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_3DLIGHT));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_3DLIGHT));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_FASTWIRE));
	if (!activate)
		w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_VIEW_AUTOROTATE));
	//w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_CMAP_BRANCH));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_CMAP_COLORBANDS));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_CMAP_DESERT));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_CMAP_GRAYSCALE));
	w[c++] = static_cast<Gtk_Widget*> (this->d_menuHash->lookup (MENU_CMAP_REDHOT));

	for (i=0; i<c; i++)
		{
		if (!w[i])
			printf ("activateMenu Error: w[%d]=NULL\n", i);
		else
			w[i]->set_sensitive (activate);
		}

	delete w;
}


//
// Overloaded event implementations...
//
gint TFWindowHandler::button_press_event_impl (GdkEventButton *e)
{
	// button 1 starts rotating 
	if (e->button == 1)
		{
		d_rotate = TRUE;
		d_rotStartPosx = (int)e->x;	// Note current x position
		d_rotStartPosy = (int)e->y;	// Note current y position
		}
	else
	// button 2 resets the view 
	if (e->button == 2)
		{
		d_rotate = FALSE;
		if (d_HFD)
			{
			d_HFD->resetView ();
			d_HFD->draw (TRUE);
			}
		}
	else
	if (e->button == 3 && !TFOptions::s_fixedMenus)
		{
		this->d_menuHandle->popup (NULL, NULL, e->button, e->time);
		}

	return Gtk_Widget::button_press_event_impl(e);
}


gint TFWindowHandler::button_release_event_impl (GdkEventButton *e)
{
	//printf ("Button Release\n"); fflush (stdout);
	// Rotation completed
	d_rotate = FALSE;

	return Gtk_Widget::button_release_event_impl(e);
}


gint TFWindowHandler::motion_notify_event_impl (GdkEventMotion *e)
{
	int 		x, y;
	GdkModifierType	state;

	//printf ("Motion Notify\n"); fflush (stdout);

	if (d_rotate && d_fastWire) 
		{
		if (e->is_hint) 
			gdk_window_get_pointer(e->window, &x, &y, &state);
		else 
			{
			x = int (e->x);
			y = int (e->y);
			//state = e->state;
			}

		//printf ("Motion Notify (%d, %d)\n", x, y); fflush (stdout);
		// Apply rotation
		d_HFD->rotateFastView (0.0, (float)((d_rotStartPosx-x)/500.0), 0.0);
		if (d_HFD->d_fastDoRotZ)
			d_HFD->rotateFastView (0.0, 0.0, (float)((y-d_rotStartPosy)/500.0));
		d_rotStartPosx = int(e->x);
		d_rotStartPosy = int(e->y);
		d_HFD->draw (FALSE);
		}
	return Gtk_Widget::motion_notify_event_impl(e);
}


